
/*
 *	Atom program to count static and dynamic occurances of
 *	instruction sequences which would be more efficient if they
 *	used the BWX instructions.  Also spot/track AMASK/IMPLVER
 *	executions as well.
 *
 * Copyright (C) 1999   Rick Gorton gorton@world.std.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <cmplrs/atom.inst.h>
#include "datatypes.h"
#include "declare.h"

#define	INT_OPERATE_MASK	0x3f

void InstrumentInit(int iargc, char **iargv)
{
  int	i,j;
  char	namebuffer[256];

  if (iargc > 1)
    strcpy(namebuffer, iargv[1]);
  else
    strcpy(namebuffer, LOGFILE_NAME_);
  Inst_Logfile = fopen(namebuffer, "a+");
  printf("Logfile will be '%s'\n", namebuffer);
  fprintf(Inst_Logfile, "Logfile:\t%s\n", namebuffer);
  fprintf(Inst_Logfile, "*** Static address data begin ***\n");

  for (i=0; i< (int)LAST_OP; i++)
    {
    static_count[i] = 0;
    }

  static_SEXTB.counter= 0x0;
  static_SEXTW.counter= 0x0;
  static_SEXTL.counter= 0x0;
  static_LDBU.counter= 0x0;
  static_LDWU.counter= 0x0;
  static_STB.counter= 0x0;
  static_STW.counter= 0x0;
  static_AMASK.counter= 0x0;
  static_IMPLVER.counter= 0x0;

  AddCallProto("SEXTB_Count()");
  AddCallProto("SEXTW_Count()");
  AddCallProto("SEXTL_Count()");
  AddCallProto("LDBU_Count()");
  AddCallProto("LDWU_Count()");
  AddCallProto("STB_Count()");
  AddCallProto("STW_Count()");
  AddCallProto("AMASK_Count()");
  AddCallProto("IMPLVER_Count()");
  AddCallProto("Total_Inst_Count(long)");

  AddCallProto("DynamicInit(char *)");
  AddCallProgram(ProgramBefore, "DynamicInit", namebuffer);
}

Instrument(int argc, char **argv, Obj *obj)
{
  Proc *p; Block *b;  Inst *this_inst;
  RegvType	RegA, RegB, RegC;
  enum Opcodes	curr_opcode;
  long		this_PC;

  for (p = GetFirstObjProc(obj); p != NULL; p = GetNextProc(p)) {
    for (b = GetFirstBlock(p); b != NULL; b = GetNextBlock(b)) {
      AddCallBlock (b, BlockBefore, "Total_Inst_Count", GetBlockInfo(b, BlockNumberInsts));
      for (this_inst = GetFirstInst(b); this_inst != NULL;
			this_inst = GetNextInst(this_inst)) {
	IClassType	this_class;

	this_class = GetInstClass(this_inst);
	RegA = GetInstRegEnum(this_inst, InstRA);
	RegB = GetInstRegEnum(this_inst, InstRB);
	RegC = GetInstRegEnum(this_inst, InstRC);
	curr_opcode = GetOpcode(this_inst);
	this_PC = InstPC(this_inst);

	switch (curr_opcode)
	    {
	    case AND_LIT_OP:
		{
		/*
		 * Look for LDBU:
		 *	ldX	Ra, x(Rbase)	Rbase known to be aligned
		 *	and	Ra, 0xff, Rb
		 */
		Inst	*ldx_op;
		enum Opcodes	tmp_opcode;

		if (GetLitValue(this_inst) != 0xff)
		    break;

		ldx_op = PrevIntRegSet(this_inst, RegA);
		if (ldx_op == NULL)
		    break;
		if (GetInstRegEnum(ldx_op, InstRA) != RegA)
		    break;
		tmp_opcode = GetOpcode(ldx_op);
		if ((tmp_opcode != LDL_OP)
			&& (tmp_opcode != LDQ_OP)
			&& (tmp_opcode != LDQU_OP))
		    break;
		fprintf(Inst_Logfile, "LDBU @ 0x%lx\n", this_PC);
		static_LDBU.counter++;
		AddCallInst(this_inst, InstAfter, "LDBU_Count");

		break;
		}

	    case SRA_LIT_OP:
		{
		/*
		 * Look for: SEXTx
		 *	sll	Ra, num, Rb
		 *	sra	Rb, num, Rc
		 * Where num is either 0x20, 0x30, or 0x38.
		 */

		Inst	*prev;
		prev = PrevIntRegSet(this_inst, RegA);
		if (prev == NULL)
		    break;
		if ((GetOpcode(prev) == SLL_LIT_OP) &&
			(GetInstRegEnum(prev, InstRC) == RegA))
		    {
		    int	lit_shift;

		    lit_shift = GetLitValue(this_inst);
		    if (lit_shift == GetLitValue(prev))
			{
			/* Got a sign extension of something. */
			if (lit_shift == 0x20)
			    {
			    fprintf(Inst_Logfile, "SEXTL @ 0x%lx\n", this_PC);
			    static_SEXTL.counter++;
			    AddCallInst(this_inst, InstAfter, "SEXTL_Count");
			    }
			else if (lit_shift == 0x30)
			    {
			    fprintf(Inst_Logfile, "SEXTW @ 0x%lx\n", this_PC);
			    static_SEXTW.counter++;
			    AddCallInst(this_inst, InstAfter, "SEXTW_Count");
			    }
			else if (lit_shift == 0x38)
			    {
			    fprintf(Inst_Logfile, "SEXTB @ 0x%lx\n", this_PC);
			    static_SEXTB.counter++;
			    AddCallInst(this_inst, InstAfter, "SEXTB_Count");
			    }
			}
		    }
		break;
		}

	    case SRL_LIT_OP:
		{
		/*
		 * Look for LDBU:
		 *	ldX	Ra, x(Rbase)	Rbase known to be aligned
		 *	srl	Ra, 0x38, Rb
		 * --------------
		 * Look for LDWU:
		 *	ldX	Ra, x(Rbase)	Rbase known to be aligned
		 *	srl	Ra, 0x30, Rb
		 */
		Inst	*ldx_op;
		enum Opcodes	tmp_opcode;
		int	this_off;

		ldx_op = PrevIntRegSet(this_inst, RegA);
		if (ldx_op == NULL)
		    break;
		if (GetInstRegEnum(ldx_op, InstRA) != RegA)
		    break;
		tmp_opcode = GetOpcode(ldx_op);
		if ((tmp_opcode != LDL_OP)
			&& (tmp_opcode != LDQ_OP)
			&& (tmp_opcode != LDQU_OP))
		    break;

		this_off = GetLitValue(this_inst);
		if (this_off == 0x38)
		    {
		    fprintf(Inst_Logfile, "LDBU @ 0x%lx\n", this_PC);
		    static_LDBU.counter++;
		    AddCallInst(this_inst, InstAfter, "LDBU_Count");
		    }
		else if (this_off == 0x30)
		    {
		    fprintf(Inst_Logfile, "LDWU @ 0x%lx\n", this_PC);
		    static_LDWU.counter++;
		    AddCallInst(this_inst, InstAfter, "LDWU_Count");
		    }

		break;
		}

	    case EXTBL_LIT_OP:
		{
		/*
		 * Look for LDBU:
		 *	ldl	Ra, x(Rbase)	Rbase known to be aligned
		 *	extbl	Ra, offset, Rc
		 */
		Inst	*ldl_op;

		ldl_op = PrevIntRegSet(this_inst, RegA);
		if (ldl_op == NULL)
		    break;
		if ((GetOpcode(ldl_op) != LDL_OP) ||
			(GetInstRegEnum(ldl_op, InstRA) != RegA))
		    break;
		fprintf(Inst_Logfile, "LDBU @ 0x%lx\n", this_PC);
		static_LDBU.counter++;
		AddCallInst(this_inst, InstAfter, "LDBU_Count");

		break;
		}

	    case EXTBL_REG_OP:
		{
		/*
		 * Look for LDBU:
		 *	ldq_u	Ra, x(Rbase)
		 *	lda	Rb, x(Rbase)
		 *	extbl	Ra, Rb, Rc
		 */
		Inst	*ldqu_op;
		Inst	*lda_op;

		lda_op = PrevIntRegSet(this_inst, RegB);
		if (lda_op == NULL)
		    break;
		if ((GetOpcode(lda_op) != LDA_OP) ||
			(GetInstRegEnum(lda_op, InstRA) != RegB))
		    break;

		ldqu_op = PrevIntRegSet(this_inst, RegA);
		if (ldqu_op == NULL)
		    break;
		if ((GetOpcode(ldqu_op) != LDQU_OP) ||
			(GetInstRegEnum(ldqu_op, InstRA) != RegA))
		    break;

		if (GetInstInfo(ldqu_op, InstMemDisp) !=
			GetInstInfo(lda_op, InstMemDisp))
		    break;

		/* found a candidate */
		fprintf(Inst_Logfile, "LDBU @ 0x%lx\n", this_PC);
		static_LDBU.counter++;
		AddCallInst(this_inst, InstAfter, "LDBU_Count");
		break;
		}

	    case EXTQH_LIT_OP:
		{
		/*
		 * Look for SEXTB:
		 *	extqh	Ra, 1, Rb
		 *	sra	Rb, 0x38, Rc
		 * Look for SEXTW:
		 *	extqh	Ra, 2, Rb
		 *	sra	Rb, 0x30, Rc
		 * Look for SEXTL:
		 *	extqh	Ra, 4, Rb
		 *	sra	Rb, 0x20, Rc
		 */

		Inst	*next;
		next = NextIntRegUse(this_inst, RegC);
		if (next == NULL)
		    break;
		if ((GetOpcode(next) == SRA_LIT_OP) &&
			(GetInstRegEnum(next, InstRA) == RegC))
		    {
		    int	lit_field;
		    int	sra_shift;

		    lit_field = GetLitValue(this_inst);
		    sra_shift = GetLitValue(next);
		    if ((lit_field == 1) && (sra_shift == 0x38))
			{
			fprintf(Inst_Logfile, "SEXTB @ 0x%lx\n", this_PC);
			static_SEXTB.counter++;
			AddCallInst(this_inst, InstAfter, "SEXTB_Count");
			}
		    else if ((lit_field == 2) && (sra_shift == 0x30))
			{
			fprintf(Inst_Logfile, "SEXTW @ 0x%lx\n", this_PC);
			static_SEXTW.counter++;
			AddCallInst(this_inst, InstAfter, "SEXTW_Count");
			}
		    else if ((lit_field == 4) && (sra_shift == 0x20))
			{
			fprintf(Inst_Logfile, "SEXTL @ 0x%lx\n", this_PC);
			static_SEXTL.counter++;
			AddCallInst(this_inst, InstAfter, "SEXTL_Count");
			}
		    }
		break;
		}

	    case EXTQH_REG_OP:
		{
		/*
		 * Look for SEXTB: (From Alpha Architecture Manual)
		 *	ldq_u	Ra, X(Rbase)
		 *	lda	Rb, X+1(Rbase)
		 *	extqh	Ra, Rb, Ra
		 *	sra	Ra, 0x38, Ra
		 * Variants:
		 *	lda	Rb, X+1(Rbase)
		 *	ldq_u	Ra, X(Rbase)
		 *	extqh	Ra, Rb, Ra
		 *	sra	Ra, 0x38, Ra
		 * ---------------------
		 *	ldq_u	Ra, X(Rbase)
		 *	lda	Rb, X+1(Rbase)
		 *	extqh	Ra, Rb, Rc
		 *	sra	Rc, 0x38, Rc
		 * ---------------------
		 */
		Inst	*ldqu_op;
		Inst	*lda_op;
		Inst	*sra_op;

		sra_op = NextIntRegUse(this_inst, RegC);
		if (sra_op == NULL)
		    break;
		if ((GetOpcode(sra_op) != SRA_LIT_OP) ||
			(GetInstRegEnum(sra_op, InstRA) != RegC) ||
			(GetLitValue(sra_op) != 0x38) ||
			(GetInstRegEnum(sra_op, InstRC) != RegC))
		    break;

		lda_op = PrevIntRegSet(this_inst, RegB);
		if (lda_op == NULL)
		    break;
		if ((GetOpcode(lda_op) != LDA_OP) ||
			(GetInstRegEnum(lda_op, InstRA) != RegB))
		    break;

		ldqu_op = PrevIntRegSet(this_inst, RegA);
		if (ldqu_op == NULL)
		    break;
		if ((GetOpcode(ldqu_op) != LDQU_OP) ||
			(GetInstRegEnum(ldqu_op, InstRA) != RegA))
		    break;

		if ((GetInstInfo(ldqu_op, InstMemDisp) + 1) !=
			GetInstInfo(lda_op, InstMemDisp))
		    break;

		/* found a candidate */
		fprintf(Inst_Logfile, "SEXTB @ 0x%lx\n", this_PC);
		static_SEXTB.counter++;
		AddCallInst(this_inst, InstAfter, "SEXTB_Count");
		break;
		}

	    case EXTWL_LIT_OP:
		{
		/*
		 * Look for LDWU:
		 *	ldX	Ra, x(Rbase)	Rbase known to be aligned
		 *	extwl	Ra, num, Rb
		 */
		Inst	*ldx_op;
		enum Opcodes	tmp_opcode;
		int	this_off;

		ldx_op = PrevIntRegSet(this_inst, RegA);
		if (ldx_op == NULL)
		    break;
		if (GetInstRegEnum(ldx_op, InstRA) != RegA)
		    break;
		tmp_opcode = GetOpcode(ldx_op);
		if ((tmp_opcode != LDL_OP)
			&& (tmp_opcode != LDQ_OP)
			&& (tmp_opcode != LDQU_OP))
		    break;

		fprintf(Inst_Logfile, "LDWU @ 0x%lx\n", this_PC);
		static_LDWU.counter++;
		AddCallInst(this_inst, InstAfter, "LDWU_Count");

		break;
		}

	    case AMASK_LIT_OP:
	    case AMASK_REG_OP:
		{
		static_AMASK.counter++;
		AddCallInst(this_inst, InstAfter, "AMASK_Count");
		break;
		}

	    case IMPLVER_OP:
		{
		static_IMPLVER.counter++;
		AddCallInst(this_inst, InstAfter, "IMPLVER_Count");
		break;
		}

	    case ZAP_LIT_OP:
		{
		/*
		 * Look for LDBU:
		 *	ldX	Ra, x(Rbase)	Rbase known to be aligned
		 *	zap	Ra, 0xfe, Rb
		 * Look for LDWU:
		 *	ldX	Ra, x(Rbase)	Rbase known to be aligned
		 *	zap	Ra, 0xfc, Rb
		 */
		Inst	*ldx_op;
		enum Opcodes	tmp_opcode;
		int	tmpoff;

		ldx_op = PrevIntRegSet(this_inst, RegA);
		if (ldx_op == NULL)
		    break;
		if (GetInstRegEnum(ldx_op, InstRA) != RegA)
		    break;
		tmp_opcode = GetOpcode(ldx_op);
		if ((tmp_opcode != LDL_OP)
			&& (tmp_opcode != LDQ_OP)
			&& (tmp_opcode != LDQU_OP))
		    break;
		tmpoff = GetLitValue(this_inst);
		if (tmpoff == 0xfe)
		    {
		    fprintf(Inst_Logfile, "LDBU @ 0x%lx\n", this_PC);
		    static_LDBU.counter++;
		    AddCallInst(this_inst, InstAfter, "LDBU_Count");
		    break;
		    }
		else if (tmpoff == 0xfc)
		    {
		    fprintf(Inst_Logfile, "LDWU @ 0x%lx\n", this_PC);
		    static_LDWU.counter++;
		    AddCallInst(this_inst, InstAfter, "LDWU_Count");
		    break;
		    }

		break;
		}

	    case ZAPNOT_LIT_OP:
		{
		/*
		 * Look for LDBU:
		 *	ldX	Ra, x(Rbase)	Rbase known to be aligned
		 *	zapnot	Ra, 0x3, Rb
		 * Look for LDWU:
		 *	ldX	Ra, x(Rbase)	Rbase known to be aligned
		 *	zapnot	Ra, 0x1, Rb
		 */
		Inst	*ldx_op;
		enum Opcodes	tmp_opcode;
		int	tmpoff;

		ldx_op = PrevIntRegSet(this_inst, RegA);
		if (ldx_op == NULL)
		    break;
		if (GetInstRegEnum(ldx_op, InstRA) != RegA)
		    break;
		tmp_opcode = GetOpcode(ldx_op);
		if ((tmp_opcode != LDL_OP)
			&& (tmp_opcode != LDQ_OP)
			&& (tmp_opcode != LDQU_OP))
		    break;
		tmpoff = GetLitValue(this_inst);
		if (tmpoff == 0x1)
		    {
		    fprintf(Inst_Logfile, "LDBU @ 0x%lx\n", this_PC);
		    static_LDBU.counter++;
		    AddCallInst(this_inst, InstAfter, "LDBU_Count");
		    break;
		    }
		else if (tmpoff == 0x3)
		    {
		    fprintf(Inst_Logfile, "LDWU @ 0x%lx\n", this_PC);
		    static_LDWU.counter++;
		    AddCallInst(this_inst, InstAfter, "LDWU_Count");
		    break;
		    }

		break;
		}

	    case MSKBL_REG_OP:
		{
		/*
		 * Look for STB:
		 *	lda	R1, X(Rbase)
		 *	ldq_u	R2, X(Rbase)
		 *	insbl	Rsrc, R1, R3
		 *	mskbl	R2, R1, R2
		 *	bis	R2, R3, R2
		 *	stq_u	R2, X(Rbase)
		 * A valid variant (not yet detected) is:
		 *	ldq_u	R2, 0(Rbase)
		 *	insbl	Rsrc, Rbase, R3
		 *	mskbl	R2, Rbase, R2
		 *	bis	R2, R3, R2
		 *	stq_u	R2, 0(Rbase)
		 * Another variant (not yet detected) is:
		 *	ldq_u	R2, X(Rbase)
		 *	lda	Rbase, X(Rbase)
		 *	insbl	Rsrc, Rbase, R3
		 *	mskbl	R2, Rbase, R2
		 *	bis	R2, R3, R2
		 *	stq_u	R2, 0(Rbase)
		 * A generalized detection for all three is:
		 *	- Check mskbl, bis, stq_u
		 *	- Pick up offset from stq_u
		 *	- Work backwards to insbl, lda/ldqu from the offset
		 */
		Inst	*lda_op;
		Inst	*ldqu_op;
		Inst	*insbl_op;
		Inst	*mskbl_op;
		Inst	*bis_op;
		Inst	*stqu_op;
		enum Opcodes	tmp_opcode;
		int	this_off;
		RegvType	R1, R2, R3, Rsrc, Rbase;

			/* This is _not_ definitive - I've probably got
			 * a couple of detection gaps here.
			 */

			/* mskbl */
		if (RegA != RegC)
		    break;

			/* bis */
		bis_op = NextIntRegUse(this_inst, RegC);
		if (bis_op == NULL)
		    break;
		if (GetOpcode(bis_op) != BIS_REG_OP)
		    break;
		R2 = GetInstRegEnum(bis_op, InstRA);
		if ((R2 != RegC) || (R2 != GetInstRegEnum(bis_op, InstRC)))
		    break;

			/* stq_u */
		stqu_op = NextIntRegUse(bis_op, R2);
		if (stqu_op == NULL)
		    break;
		if (GetOpcode(stqu_op) != STQU_OP)
		    break;
		this_off = GetInstInfo(stqu_op, InstMemDisp);
		Rbase = GetInstRegEnum(stqu_op, InstRB);

			/* Step back for insbl */
		insbl_op = PrevIntRegUse(this_inst, RegB);
		if (insbl_op == NULL)
		    break;
		if (GetOpcode(insbl_op) != INSBL_REG_OP)
		    break;
		Rsrc = GetInstRegEnum(insbl_op, InstRA);
		R3 = GetInstRegEnum(insbl_op, InstRC);
		if (GetInstRegEnum(insbl_op, InstRB) != RegB)
		    break;
			/* Confirm insbl <---> bis connnection */
		if (PrevIntRegSet(bis_op, GetInstRegEnum(bis_op, InstRB)) != insbl_op)
		    break;

		if (this_off != 0)
		    {
			/*
			 * First variant
			 * R1 <--- RegB
			 * need: Rbase != R1
			 */
		    R1 = RegB;
		    if (Rbase == R1)
			break;
		    if (GetInstInfo(insbl_op, InstRB) != R1)
			break;
			/* lda */
		    lda_op = PrevIntRegSet(this_inst, RegB);
		    if (lda_op == NULL)
			break;
		    if (GetOpcode(lda_op) != LDA_OP)
			break;
		    if (this_off != GetInstInfo(lda_op, InstMemDisp))
			break;
		    if (Rbase != GetInstRegEnum(lda_op, InstRB))
			break;
			/* ldqu */
		    ldqu_op = PrevIntRegSet(this_inst, RegA);
		    if (ldqu_op == NULL)
			break;
		    if (GetOpcode(ldqu_op) != LDQU_OP)
			break;
		    if (this_off != GetInstInfo(ldqu_op, InstMemDisp))
			break;
		    if (Rbase != GetInstRegEnum(ldqu_op, InstRB))
			break;
			/* Got a match */
		    }
		else {	/* zero offset (no lda), or fewer registers used */
		    ldqu_op = PrevIntRegSet(this_inst, RegA);
		    if (ldqu_op == NULL)
			break;
		    if (GetOpcode(ldqu_op) != LDQU_OP)
			break;
		    if (GetInstRegEnum(ldqu_op, InstRB) != Rbase)
			break;
		    this_off = GetInstInfo(ldqu_op, InstMemDisp);
		    if (this_off != 0)
			{	/* Look at the lda */
			lda_op = PrevIntRegSet(this_inst, RegB);
			if (lda_op == NULL)
			    break;
			if (GetOpcode(lda_op) != LDA_OP)
			    break;
			if (GetInstRegEnum(lda_op, InstRB) != Rbase)
			    break;
			if (this_off != GetInstInfo(lda_op, InstMemDisp))
			    break;
			}
		    }

			/* Major tedium done (I think) */
		fprintf(Inst_Logfile, "STB @ 0x%lx\n", this_PC);
		static_STB.counter++;
		AddCallInst(this_inst, InstAfter, "STB_Count");
		break;
		}

	    case INSWL_LIT_OP:
		{
		/*
		 * NOTE: This one is a guess.
		 * Look for probable STW:
		 *	inswl	Rsrc, 0 or 4, R1
		 *	bis	R2, R1, R3
		 *	stq	R3, off(Rbase)
		 * This seems to happen a lot as part of a longer sequence:
		 *	ldq	R1, off(Rbase)
		 *	extwl	R2, 4, R3
		 *	zapnot	R2, 0xcf, R4
		 *	<bit fiddle of R3>
		 *	inswl	R3, 4, R5
		 *	bis	R4, R5, R2
		 *	stq	R2, off(Rbase)
		 */
		Inst	*bis_op;
		Inst	*stq_op;
		enum Opcodes	tmp_opcode;
		int	this_off;
		RegvType	R1;

		this_off = GetLitValue(this_inst);
		if ((this_off != 0x0) && (this_off != 0x4))
		    break;

			/* bis */
		bis_op = NextIntRegUse(this_inst, RegC);
		if (bis_op == NULL)
		    break;
		if (GetOpcode(bis_op) != BIS_REG_OP)
		    break;
		R1 = GetInstRegEnum(bis_op, InstRB);
		if (R1 != RegC)
		    break;

			/* stq_u */
		stq_op = NextIntRegUse(bis_op, GetInstRegEnum(bis_op, InstRC));
		if (stq_op == NULL)
		    break;
		if (GetOpcode(stq_op) != STQ_OP)
		    break;

		fprintf(Inst_Logfile, "STW @ 0x%lx\n", this_PC);
		static_STW.counter++;
		AddCallInst(this_inst, InstAfter, "STW_Count");
		break;
		}

	    case INSWL_REG_OP:
		{
		/*
		 * Look for STW:
		 *	bic	Rorig, 2, Rbase
		 *	ldl	R1, off(Rbase)
		 *	and	Rorig, 0x2, R2
		 *	inswl	Rsrc, R2, R3
		 *	mskwl	R1, R2, R1
		 *	bis	R1, R3, R1
		 *	stl	R1, off(Rbase)
		 */
		Inst	*bic_op;
		Inst	*ldl_op;
		Inst	*and_op;
		Inst	*mskwl_op;
		Inst	*bis_op;
		Inst	*stl_op;
		enum Opcodes	tmp_opcode;
		int	this_off;
		RegvType	Rorig, Rbase, R1, R2, R3;

			/* inswl */
			/* Rsrc == R3 is valid */
		if ((RegA == RegB) || (RegB == RegC))
		    break;

			/* and */
		and_op = PrevIntRegSet(this_inst, RegB);
		if (and_op == NULL)
		    break;
		if (GetOpcode(and_op) != AND_LIT_OP)
		    break;
		if (GetLitValue(and_op) != 0x2)
		    break;
		Rorig = GetInstRegEnum(and_op, InstRA);

			/* bic */
		bic_op = PrevIntRegUse(and_op, Rorig);
		if (bic_op == NULL)
		    break;
		if (GetOpcode(bic_op) != BIC_LIT_OP)
		    break;
		if (GetLitValue(bic_op) != 0x2)
		    break;
		Rbase = GetInstRegEnum(bic_op, InstRC);

			/* ldl */
		ldl_op = NextIntRegUse(bic_op, Rbase);
		if (ldl_op == NULL)
		    break;
		if (GetOpcode(ldl_op) != LDL_OP)
		    break;
		this_off = GetInstInfo(ldl_op, InstMemDisp);
		R1 = GetInstRegEnum(ldl_op, InstRA);

			/* mskwl */
		mskwl_op = NextIntRegUse(ldl_op, R1);
		if (mskwl_op == NULL)
		    break;
		if (GetOpcode(mskwl_op) != MSKWL_REG_OP)
		    break;
		if (R1 != GetInstRegEnum(mskwl_op, InstRA))
		    break;
		R2 = GetInstRegEnum(mskwl_op, InstRB);
		if (R2 != RegB)
		    break;
		if (PrevIntRegUse(mskwl_op, R2) != this_inst)
		    break;
		if (R1 != GetInstRegEnum(mskwl_op, InstRC))
		    break;

			/* bis */
		bis_op = NextIntRegUse(mskwl_op, R1);
		if (bis_op == NULL)
		    break;
		if (GetOpcode(bis_op) != BIS_REG_OP)
		    break;
		R3 = GetInstRegEnum(bis_op, InstRB);
		if (R3 != RegC)
		    break;
		if (PrevIntRegSet(bis_op, R3) != this_inst)
		    break;
		if (GetInstRegEnum(bis_op, InstRA) != GetInstRegEnum(bis_op, InstRC))
		    break;

			/* stl */
		stl_op = NextIntRegUse(bis_op, R1);
		if (ldl_op == NULL)
		    break;
		if (GetOpcode(ldl_op) != LDL_OP)
		    break;
		if (GetInstRegEnum(ldl_op, InstRA) != R1)
		    break;
		if (this_off != GetInstInfo(ldl_op, InstMemDisp))
		    break;
		if (Rbase != GetInstRegEnum(ldl_op, InstRB))
		    break;

		fprintf(Inst_Logfile, "STW @ 0x%lx\n", this_PC);
		static_STW.counter++;
		AddCallInst(this_inst, InstAfter, "STW_Count");

		break;
		}

	    default:
		break;	/* All other opcodes */
	    }	/* switch on relevant opcodes */
	}	/* Each Inst in a block */
      }	/* Each block in a proc */
    }	/* Each proc */
}

void InstrumentFini()
{
  int	i;

  fprintf(Inst_Logfile, "*** Static data Summary begin ***\n");
  fprintf(Inst_Logfile, "    Potential BWX candidates\n");
  fprintf(Inst_Logfile, "\tSEXTB candidates:\t%ld\n", static_SEXTB.counter);
  fprintf(Inst_Logfile, "\tSEXTW candidates:\t%ld\n", static_SEXTW.counter);
  fprintf(Inst_Logfile, "\tSEXTL candidates (ADDL):\t%ld\n", static_SEXTL.counter);
  fprintf(Inst_Logfile, "\tLDBU candidates:\t%ld\n", static_LDBU.counter);
  fprintf(Inst_Logfile, "\tLDWU candidates:\t%ld\n", static_LDWU.counter);
  fprintf(Inst_Logfile, "\tSTB candidates:\t%ld\n", static_STB.counter);
  fprintf(Inst_Logfile, "\tSTW candidates:\t%ld\n", static_STW.counter);
  fprintf(Inst_Logfile, "    AMASK and IMPLVER uses\n");
  fprintf(Inst_Logfile, "\tAMASK uses:\t%ld\n", static_AMASK.counter);
  fprintf(Inst_Logfile, "\tIMPLVER uses:\t%ld\n", static_IMPLVER.counter);
  fprintf(Inst_Logfile, "*** Static data Summary end ***\n");
  fclose(Inst_Logfile);

  AddCallProto("DynamicFini()");
  AddCallProgram(ProgramAfter, "DynamicFini");
}
